@@ -10,6 +10,7 @@
 #include "factor.h"
 
 #define FUNC_gcd_ui 1
+#define FUNC_mpz_logn 1
 #include "utility.h"
 
 static mpz_t _bgcd;
@@ -153,7 +154,7 @@ int primality_pretest(mpz_t n)
   (log2n < 100) ? 1000 : \
   (BITS_PER_WORD == 32 && log2n > 9000U) ? UVCONST(2500000000) : \
   (BITS_PER_WORD == 64 && log2n > 4294967294U) ? UVCONST(9300000000000000000) :\
-  ((log2n * (log2n >> 5) * log2log2n) >> 1)
+  ((log2n * (log2n >> 5) * (UV)((log2log2n)*1.5)) >> 1)
 
 static void next_prime_with_sieve(mpz_t n) {
   UV i, log2n, log2log2n, width, depth;
@@ -279,6 +280,91 @@ void _GMP_prev_prime(mpz_t n)
   }
 }
 
+void surround_primes(mpz_t n, UV* prev, UV* next, UV skip_width) {
+  UV i, j, log2n, log2log2n, width, depth, fprev, fnext, search_merits;
+  uint32_t* comp;
+  mpz_t t, base;
+  int neven, found;
+
+  log2n = mpz_sizeinbase(n, 2);
+  for (log2log2n = 1, i = log2n; i >>= 1; ) log2log2n++;
+
+  if (log2n < 64) {
+    mpz_init(t);
+    mpz_set(t, n);
+    _GMP_prev_prime(t);
+    mpz_sub(t, n, t);
+    *prev = mpz_get_ui(t);
+    mpz_set(t, n);
+    _GMP_next_prime(t);
+    mpz_sub(t, t, n);
+    *next = mpz_get_ui(t);
+    mpz_clear(t);
+    return;
+  }
+
+  mpz_init(t);
+  mpz_init(base);
+  fprev = fnext = 0;
+  neven = mpz_even_p(n);
+  j = 1 + !neven;         /* distance from n we're looking. */
+
+  for (found = 0, search_merits = 15; !found; search_merits *= 2) {
+    double logn = mpz_logn(n);
+
+    if (BITS_PER_WORD == 32 && log2n >   7000)
+      depth = UVCONST(2500000000);
+    else if (BITS_PER_WORD == 64 && log2n > 200000)
+      depth = UVCONST(6000000000000);
+    else if (log2n > 900)
+      depth = (UV) ((-.05L+(log2n/8000.0L)) * logn * logn * log(logn));
+    else
+      depth = NPS_DEPTH(log2n, log2log2n);
+
+    width = (UV) (search_merits * logn + 0.5);
+    width = 64 * ((width+63)/64);    /* Round up to next 64 */
+    if (neven) width++;              /* base will always be odd */
+    mpz_sub_ui(base, n, width);
+    /* printf("merits %lu  width %lu  depth %lu  skip_width %lu\n", search_merits, width, depth, skip_width); */
+
+    /* gmp_printf("partial sieve width %lu  depth %lu\n", 2*width+1, depth); */
+    comp = partial_sieve(base, 2*width+1, depth);
+
+    for (; j < width; j += 2) {
+      if (!fprev) {
+        if (!TSTAVAL(comp, width+1-j)) {
+          mpz_sub_ui(t, n, j);
+          if ( (skip_width == 0) ? _GMP_BPSW(t) : miller_rabin_ui(t,2) ) {
+            fprev = j;
+            if (fnext || (skip_width != 0 && j <= skip_width))
+              break;
+          }
+        }
+      }
+      if (!fnext) {
+        if (!TSTAVAL(comp, width+1+j)) {
+          mpz_add_ui(t, n, j);
+          if ( (skip_width == 0) ? _GMP_BPSW(t) : miller_rabin_ui(t,2) ) {
+            fnext = j;
+            if (fprev || (skip_width != 0 && j <= skip_width))
+              break;
+          }
+        }
+      }
+    }
+
+    Safefree(comp);
+    if ( (fprev && fnext) ||
+         (skip_width != 0 && j <= skip_width && (fprev || fnext)) )
+      found = 1;
+  }
+
+  mpz_clear(base);
+  mpz_clear(t);
+
+  *prev = fprev;
+  *next = fnext;
+}
 
 /*****************************************************************************/
 
@@ -390,6 +476,8 @@ void bernfrac(mpz_t num, mpz_t den, mpz_t zn)
   else if (n == 1) { mpz_set_ui(num, 1);  mpz_set_ui(den, 2); return; }
   else if (n & 1)  { mpz_set_ui(num, 0);  mpz_set_ui(den, 1); return; }
   n >>= 1;
+
+  /* Algorithm TangentNumbers from https://arxiv.org/pdf/1108.0286.pdf */
   New(0, T, n+1, mpz_t);
   for (k = 1; k <= n; k++)  mpz_init(T[k]);
   mpz_set_ui(T[1], 1);
@@ -407,6 +495,7 @@ void bernfrac(mpz_t num, mpz_t den, mpz_t zn)
     }
   }
 
+  /* (14), also last line of Algorithm FastTangentNumbers from paper */
   mpz_mul_ui(num, T[n], n);
   mpz_mul_si(num, num, (n & 1) ? 2 : -2);
   mpz_set_ui(t, 1);
@@ -416,6 +505,7 @@ void bernfrac(mpz_t num, mpz_t den, mpz_t zn)
   mpz_gcd(t, num, den);
   mpz_divexact(num, num, t);
   mpz_divexact(den, den, t);
+
   mpz_clear(t);
   for (k = 1; k <= n; k++)  mpz_clear(T[k]);
   Safefree(T);
@@ -453,34 +543,51 @@ void harmfrac(mpz_t num, mpz_t den, mpz_t zn)
   mpz_clear(t);
 }
 
-char* harmreal(mpz_t zn, unsigned long prec) {
+static char* frac_real(mpz_t num, mpz_t den, unsigned long prec) {
   char* out;
-  mpz_t num, den;
   mpf_t fnum, fden, res;
-  unsigned long numsize, densize;
+  unsigned long numbits = mpz_sizeinbase(num,  2);
+  unsigned long denbits = mpz_sizeinbase(den,  2);
+  unsigned long numdigs = mpz_sizeinbase(num, 10);
+  unsigned long dendigs = mpz_sizeinbase(den, 10);
 
-  mpz_init(num); mpz_init(den);
-  harmfrac(num, den, zn);
-
-  numsize = mpz_sizeinbase(num, 10);
-  densize = mpz_sizeinbase(den, 10);
-
-  mpf_init2(fnum, 1 + mpz_sizeinbase(num,2));
-  mpf_init2(fden, 1 + mpz_sizeinbase(den,2));
-  mpf_set_z(fnum, num);  mpf_set_z(fden, den);
-  mpz_clear(den); mpz_clear(num);
-
-  mpf_init2(res, (unsigned long) (8+prec*3.4) );
+  mpf_init2(fnum, 1 + numbits);  mpf_set_z(fnum, num);
+  mpf_init2(fden, 1 + denbits);  mpf_set_z(fden, den);
+  mpf_init2(res, (unsigned long) (8 + (numbits-denbits+1) + prec*3.4) );
   mpf_div(res, fnum, fden);
   mpf_clear(fnum);  mpf_clear(fden);
 
-  New(0, out, (10+numsize-densize)+prec, char);
+  New(0, out, (10+numdigs-dendigs)+prec, char);
   gmp_sprintf(out, "%.*Ff", (int)(prec), res);
   mpf_clear(res);
 
   return out;
 }
 
+char* harmreal(mpz_t zn, unsigned long prec) {
+  char* out;
+  mpz_t num, den;
+
+  mpz_init(num); mpz_init(den);
+  harmfrac(num, den, zn);
+  out = frac_real(num, den, prec);
+  mpz_clear(den); mpz_clear(num);
+
+  return out;
+}
+
+char* bernreal(mpz_t zn, unsigned long prec) {
+  char* out;
+  mpz_t num, den;
+
+  mpz_init(num); mpz_init(den);
+  bernfrac(num, den, zn);
+  out = frac_real(num, den, prec);
+  mpz_clear(den); mpz_clear(num);
+
+  return out;
+}
+
 void stirling(mpz_t r, unsigned long n, unsigned long m, UV type)
 {
   mpz_t t, t2;
