@@ -5,7 +5,7 @@ use Carp qw/croak confess carp/;
 
 BEGIN {
   $Math::Prime::Util::GMP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::GMP::VERSION = '0.38';
+  $Math::Prime::Util::GMP::VERSION = '0.39';
 }
 
 # parent is cleaner, and in the Perl 5.10.1 / 5.12.0 core, but not earlier.
@@ -23,6 +23,7 @@ our @EXPORT_OK = qw(
                      is_bls75_prime
                      is_ecpp_prime
                      is_pseudoprime
+                     is_euler_plumb_pseudoprime
                      is_strong_pseudoprime
                      is_lucas_pseudoprime
                      is_strong_lucas_pseudoprime
@@ -45,6 +46,7 @@ our @EXPORT_OK = qw(
                      sieve_range
                      next_prime
                      prev_prime
+                     surround_primes
                      trial_factor
                      prho_factor
                      pbrent_factor
@@ -63,7 +65,7 @@ our @EXPORT_OK = qw(
                      pn_primorial
                      factorial
                      consecutive_integer_lcm
-                     partitions bernfrac harmfrac harmreal stirling
+                     partitions bernfrac bernreal harmfrac harmreal stirling
                      gcd lcm kronecker valuation binomial gcdext
                      invmod sqrtmod addmod mulmod divmod powmod
                      vecsum vecprod
@@ -171,7 +173,7 @@ __END__
 
 =encoding utf8
 
-=for stopwords Möbius Deléglise Bézout gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac harmfrac harmreal stirling lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod
+=for stopwords Möbius Deléglise Bézout gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac bernreal harmfrac harmreal stirling lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod
 
 =head1 NAME
 
@@ -180,7 +182,7 @@ Math::Prime::Util::GMP - Utilities related to prime numbers and factoring, using
 
 =head1 VERSION
 
-Version 0.38
+Version 0.39
 
 
 =head1 SYNOPSIS
@@ -542,6 +544,11 @@ of the other probable prime tests and offers little benefit, especially over
 combined tests like L</is_bpsw_prime> and
 L</is_frobenius_underwood_pseudoprime>.
 
+An optional second argument indicates whether to additionally test C<P(-n)>.
+Composites which pass this test are known as restricted Perrin pseudoprimes
+and are a subset of the unrestricted set.
+
+
 =head2 is_frobenius_pseudoprime
 
 Takes a positive number C<n> as input, and two optional parameters C<a> and
@@ -558,7 +565,7 @@ efficient Frobenius test of Paul Underwood.  This selects a parameter C<a>
 as the least non-negative integer such that C<(a^2-4|n)=-1>, then verifies that
 C<(x+2)^(n+1) = 2a + 5 mod (x^2-ax+1,n)>.  This combines a Fermat and Lucas
 test at a computational cost of about 2.5x a strong pseudoprime test.  This
-makes it similar to, but faster than, a Frobenius test.
+makes it similar to, but faster than, a standard Frobenius test.
 
 This test is deterministic (no randomness is used).  There are no known
 pseudoprimes to this test.  This test also has no overlap with the BPSW
@@ -850,6 +857,26 @@ The function L</is_prob_prime> is used to determine when a prime is found,
 hence the result is a probable prime (using BPSW).
 
 
+=head2 surround_primes
+
+  ($dprev, $dnext) = surround_primes($n);
+
+Returns the distances to the previous and next primes of the input C<n>.
+This is slightly more efficient than calling both L</prev_prime> and
+L</next_prime>, and returning the native integer distances can be more
+efficient with large inputs.
+
+If an optional second argument C<d> is given, and the input C<n> is larger
+than C<2^64>, then if a SPSP-2 is found in the range C<n-d> to C<n+d>
+(inclusive) then it will be returned with the other argument set to C<0>.
+Otherwise, the first SPSP-2 values found are returned.
+This feature is especially useful for prime gap searches as well as
+finding the nearest prime to a value.
+
+Note that with a non-zero second argument, the values returned have not
+undergone a full BPSW test; just sieving and a SPSP-2 test.
+
+
 =head2 lucasu
 
   say "Fibonacci($_) = ", lucasu(1,-1,$_) for 0..100;
@@ -997,6 +1024,12 @@ B_1 = 1/2.
 This corresponds to Pari's C<bernfrac(n)> and Mathematica's C<BernoulliB>
 functions.
 
+=head2 bernreal
+
+Returns the Bernoulli number C<B_n> for an integer argument C<n>, as a
+string floating point.  An optional second argument indicates the number
+of digits to be preserved past the decimal place, with a default of 40.
+
 =head2 harmfrac
 
 Returns the Harmonic number C<H_n> for an integer argument C<n>, as a
