@@ -4,15 +4,18 @@
 #include <gmp.h>
 #include "ptypes.h"
 
-extern int  _GMP_miller_rabin(mpz_t n, mpz_t a);
+extern int  miller_rabin(mpz_t n, mpz_t a);
+extern int  miller_rabin_ui(mpz_t n, unsigned long a);
+extern int  miller_rabin_random(mpz_t n, UV numbases, char* seedstr);
+
 extern int  _GMP_is_lucas_pseudoprime(mpz_t n, int strength);
 extern int  _GMP_is_almost_extra_strong_lucas_pseudoprime(mpz_t n, UV incr);
 extern int  _GMP_is_frobenius_underwood_pseudoprime(mpz_t n);
 extern int  _GMP_is_frobenius_khashin_pseudoprime(mpz_t n);
-extern int  is_perrin_pseudoprime(mpz_t n);
+extern int  is_perrin_pseudoprime(mpz_t n, int restricted);
+extern int  is_euler_plumb_pseudoprime(mpz_t n);
 extern int  is_frobenius_pseudoprime(mpz_t n, IV P, IV Q);
 extern int  is_frobenius_cp_pseudoprime(mpz_t n, UV ntests);
-extern int  _GMP_miller_rabin_random(mpz_t n, UV numbases, char* seedstr);
 
 extern void lucas_seq(mpz_t U, mpz_t V, mpz_t n, IV P, IV Q, mpz_t k,
                       mpz_t Qk, mpz_t t);
